SystemOrganization addCategory: #'Signals-Tests'!
SystemOrganization addCategory: #'Signals-Core'!

TestCase subclass: #SignalTest
	instanceVariableNames: 'producer consumer process'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Signals-Tests'!

!SignalTest methodsFor: 'running' stamp: 'mt 6/2/2010 22:42'!
setUp

	consumer := SignalConsumer new.
	producer := SignalProducer new.! !

!SignalTest methodsFor: 'running' stamp: 'mt 6/3/2010 08:19'!
tearDown

	super tearDown.
	(process notNil and: [process isTerminated not]) ifTrue: [process terminate].! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:00'!
test01ConnectDisconnect

	self assert: producer signalConnections isEmpty.
	consumer connect: producer signal: #valueChanged: toSelector: #value:.
	self assert: 1 equals: producer signalConnections size.
	
	self
		assert: (producer disconnectSignal: #valueChanged:);
		assert: producer signalConnections isEmpty;
		deny: (producer disconnectSignal: #valueChanged:).
		
	self connect: producer signal: #valueChanged: to: consumer selector: #value:.
	self assert: 1 equals: producer signalConnections size.

	self
		assert: (producer disconnectSignal: #valueChanged: from: consumer);
		assert: producer signalConnections isEmpty;
		deny: (producer disconnectSignal: #valueChanged: from: consumer).
		
	self connect: producer signal: #valueChanged: to: consumer selector: #value:.
	
	self
		assert: (producer disconnectSignal: #valueChanged: from: consumer selector: #value:);
		assert: producer signalConnections isEmpty;
		deny: (producer disconnectSignal: #valueChanged: from: consumer selector: #value:).
		
	self connect: producer signal: #valueChanged: to: consumer selector: #value:.
	self 
		deny: (producer disconnectSignal: #shouldNotExist);
		deny: producer signalConnections isEmpty.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:00'!
test02ConnectError

	self
		should: [consumer connect: producer signal: #shouldNotExist toSelector: #toggleSwitch] raise: Error;
		should: [consumer connect: producer signal: #valueChanged: toSelector: #shouldNotExist] raise: Error;
		should: [self connect: producer signal: #shouldNotExist to: consumer selector: #toggleSwitch] raise: Error;
		should: [self connect: producer signal: #valueChanged: to: consumer selector: #shouldNotExist] raise: Error.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:20'!
test03Signaling

	consumer connect: producer signal: #valueChanged: toSelector: #value:.
	consumer connect: producer signal: #newSignalProduced toSelector: #toggleSwitch.
	
	self
		assert: consumer value isNil;
		assert: #off equals: consumer switchState.	
	producer valueChanged: 'foobar'.
	producer newSignalProduced.
	self
		assert: consumer value = 'foobar';
		assert: #on equals: consumer switchState.	
	
	producer disconnectAllSignals.
	producer valueChanged: 'blubb'.
	self
		assert: consumer value = 'foobar';
		assert: #on equals: consumer switchState.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:17'!
test04Patterns

	consumer connect: producer signal: #valueChanged: toSelector: #toggleSwitch pattern: #().
	
	self assert:#off equals: consumer switchState.	
	producer valueChanged: 'foobar'.
	self assert:#on equals: consumer switchState.	! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:21'!
test05PatternsReorder

	consumer connect: producer signal: #valueChanged:and: toSelector: #value: pattern: #(2).
	
	self assert: consumer value isNil.
	producer valueChanged: 'foobar' and: 'blubb'.
	self assert: 'blubb' equals: consumer value.	! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/2/2010 22:22'!
test06DoubleConnection

	consumer connect: producer signal: #valueChanged: toSelector: #value:.
	consumer connect: producer signal: #valueChanged: toSelector: #toggleSwitch pattern: #().
	
	self
		assert: consumer value isNil;
		assert: #off equals: consumer switchState.	
	
	producer valueChanged: 'foobar'.
	
	self
		assert: consumer value = 'foobar';
		assert: #on equals: consumer switchState.	! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/3/2010 08:13'!
test07PatternError

	self
		should: [consumer connect: producer signal: #valueChanged: toSelector: #toggleSwitch pattern: #(1)] raise: Error;
		should: [consumer connect: producer signal: #valueChanged: toSelector: #value: pattern: #(2)] raise: Error;
		should: [consumer connect: producer signal: #valueChanged:with: toSelector: #value: pattern: #(2 1)] raise: Error.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/3/2010 08:20'!
test08QueuedConnection

	| queue result |
	queue := SharedQueue new.
	process := [queue next value. result := consumer value] forkAt: 60.
	
	consumer connect: producer signal: #valueChanged: toSelector: #value: queue: queue.
	producer valueChanged: #QueuedConnectionsAreCool.
	
	Processor yield. "Let the result be set."
	
	self assert: #QueuedConnectionsAreCool equals: result.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/3/2010 08:28'!
test09QueuedConnectionToUI

	consumer
		connect: producer
		signal: #valueChanged:
		toSelector: #value:
		queue: WorldState deferredUIMessages.
		
	producer valueChanged: #QueuedConnectionsAreCool.
	
	self assert: nil equals: consumer value.
	World doOneCycle. "Let the result be set. Do it manually here because we ARE already in the UI process."
	self assert: #QueuedConnectionsAreCool equals: consumer value.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/3/2010 10:38'!
test10BlockingConnection

	| result |
	consumer
		connect: producer
		signal: #valueChanged:
		toSelector: #value:
		queue: WorldState deferredUIMessages
		blocking: true.
		
	process := [producer valueChanged: #QueuedConnectionsAreCool. result := 'blubb'] forkAt: 60.
	Processor yield.
	
	self assert: nil equals: result.
	
	World doOneCycle. "Deliver the signal."
	Processor yield. "Let the result be set."
	
	self assert: 'blubb' equals: result.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/3/2010 12:24'!
test11SignalSpy

	| spy |
	spy := SignalSpy onSender: producer signal: #valueChanged:.
	producer valueChanged: #DontSpyOnMe.	
	self assert: {#valueChanged:. {#DontSpyOnMe}} equals: spy waitForNextSignal.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/3/2010 12:24'!
test12SignalSpyConcurrent

	| spy |
	spy := SignalSpy onSender: producer signal: #valueChanged:.
	process := [producer valueChanged: #DontSpyOnMe] fork.
	self assert: {#valueChanged:. {#DontSpyOnMe}} equals: spy waitForNextSignal.! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/3/2010 12:24'!
test13SignalSpySpecificSignal

	| spy |
	spy := SignalSpy onSender: producer signal: #valueChanged:.
	process := [
		producer newSignalProduced. "will be discarded"
		producer valueChanged: #DontSpyOnMe] fork.
	self assert: {#valueChanged:. {#DontSpyOnMe}} equals: (spy waitForSignal: #valueChanged:).! !

!SignalTest methodsFor: 'tests' stamp: 'mt 6/4/2010 13:06'!
test14DefaultUIProcessing
	"Check whether default connections out of the ui process will only be processed in the ui process."
	
	self assert: Project uiProcess equals: Processor activeProcess.
	
	consumer
		connect: producer
		signal: #valueChanged:
		toSelector: #value:.
		
	process := [producer valueChanged: #QueuedConnectionsAreCool] forkAt: 60.
	Processor yield.
	
	self assert: nil equals: consumer value.
	
	World doOneCycle. "Deliver the signal."
	
	self assert: #QueuedConnectionsAreCool equals: consumer value.! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:23'!
connect: aSender signal: aSignal to: aReceiver selector: aSelector
	"Mediator: Connect a signaling object with a receiver."

	aSender
		connectSignal: aSignal
		to: aReceiver
		selector: aSelector
		pattern: nil
		queue: nil
		blocking: false! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:23'!
connect: aSender signal: aSignal to: aReceiver selector: aSelector pattern: argPattern
	"Shadow or reorder arguments before invoking selector on receiver, e.g.,  #(1 4 3)"

	aSender
		connectSignal: aSignal
		to: aReceiver
		selector: aSelector
		pattern: argPattern
		queue: nil
		blocking: false! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:23'!
connect: aSender signal: aSignal to: aReceiver selector: aSelector pattern: argPattern queue: aSharedQueue
	"Shadow or reorder arguments before invoking selector on receiver, e.g.,  #(1 4 3)"

	aSender
		connectSignal: aSignal
		to: aReceiver
		selector: aSelector
		pattern: argPattern
		queue: aSharedQueue
		blocking: false! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:17'!
connect: aSender signal: aSignal to: aReceiver selector: aSelector pattern: argPattern queue: aSharedQueue blocking: aBoolean
	"Shadow or reorder arguments before invoking selector on receiver, e.g.,  #(1 4 3)"

	aSender
		connectSignal: aSignal
		to: aReceiver
		selector: aSelector
		pattern: argPattern
		queue: aSharedQueue
		blocking: aBoolean! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:25'!
connect: aSender signal: aSignal to: aReceiver selector: aSelector queue: aSharedQueue
	"Store an incoming signal connection request. This is a direct connection, so the receiver will be called synchonously after the signal was emitted. See #emit."
	
	aSender
		connectSignal: aSignal
		to: aReceiver
		selector: aSelector
		pattern: nil
		queue: aSharedQueue
		blocking: false! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:18'!
connect: aSender signal: aSignal to: aReceiver selector: aSelector queue: aSharedQueue blocking: aBoolean

	aSender
		connectSignal: aSignal
		to: aReceiver
		selector: aSelector
		pattern: nil
		queue: aSharedQueue
		blocking: aBoolean! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:23'!
connect: aSender signal: aSignal toSelector: aSelector
	"Connect a selector from me with the signal of a sender."

	aSender
		connectSignal: aSignal
		to: self
		selector: aSelector
		pattern: nil
		queue: nil
		blocking: false.! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:24'!
connect: aSender signal: aSignal toSelector: aSelector pattern: argPattern
	"Connect a selector from me with the signal of a sender."

	aSender
		connectSignal: aSignal
		to: self
		selector: aSelector
		pattern: argPattern
		queue: nil
		blocking: false.! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:24'!
connect: aSender signal: aSignal toSelector: aSelector pattern: argPattern queue: aSharedQueue
	"Create a queued connection. Receiver will not be called directly but a valuable block with the call will be stored into a queue. The queue has to be processed somewhere else. See #emit."

	aSender
		connectSignal: aSignal
		to: self
		selector: aSelector
		pattern: argPattern
		queue: aSharedQueue
		blocking: false.! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:20'!
connect: aSender signal: aSignal toSelector: aSelector pattern: argPattern queue: aSharedQueue blocking: aBoolean
	"Create a queued connection. Receiver will not be called directly but a valuable block with the call will be stored into a queue. The queue has to be processed somewhere else. See #emit."

	aSender
		connectSignal: aSignal
		to: self
		selector: aSelector
		pattern: argPattern
		queue: aSharedQueue
		blocking: aBoolean.! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:24'!
connect: aSender signal: aSignal toSelector: aSelector queue: aSharedQueue
	"Connect a selector from me with the signal of a sender."

	aSender
		connectSignal: aSignal
		to: self
		selector: aSelector
		pattern: nil
		queue: aSharedQueue
		blocking: false.! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/3/2010 10:20'!
connect: aSender signal: aSignal toSelector: aSelector queue: aSharedQueue blocking: aBoolean
	"Connect a selector from me with the signal of a sender."

	aSender
		connectSignal: aSignal
		to: self
		selector: aSelector
		pattern: nil
		queue: aSharedQueue
		blocking: aBoolean.! !

!Object methodsFor: '*signals-core-private' stamp: 'mt 6/4/2010 12:51'!
connectSignal: aSignal to: aReceiver selector: aSelector pattern: argPattern queue: aSharedQueue blocking: aBoolean
	"Private. "

	| forceUiProcess |
	((self respondsTo: aSignal)
		and: [aReceiver respondsTo: aSelector])
		ifFalse: [Error signal: 'Could not connect sender to receiver!!'].
	
	(argPattern notNil and: [argPattern notEmpty]) ifTrue: [
		(argPattern min <= 0
			or: [(aReceiver class >> aSelector) numArgs ~= argPattern size]
			or: [(self class >> aSignal) numArgs < argPattern max])
			ifTrue: [Error signal: 'Could not connect sender to receiver!! Pattern is invalid!!']].

	"Signals sent in a non-multiprogramming environment should be processed in the UI process no matter from which process they are emitted."
	forceUiProcess := Processor activeProcess == Project uiProcess and: [aSharedQueue isNil].
	SignalConnectionsRepository critical: [
		((self signalConnections
			at: aSignal
			ifAbsentPut: [WeakIdentityKeyDictionary new])
			at: aReceiver
			ifAbsentPut: [OrderedCollection new])
			add: aSelector -> {argPattern. aSharedQueue. aBoolean. forceUiProcess}].! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 22:41'!
disconnectAllSignals
	
	SignalConnectionsRepository instance
		removeKey: self
		ifAbsent: [^ false].
		
	^ true! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 21:35'!
disconnectSignal: aSignal
	
	self signalConnections
		removeKey: aSignal
		ifAbsent: [^ false].
		
	^ true! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 21:35'!
disconnectSignal: aSignal from: aReceiver
	
	(self signalConnections includesKey: aSignal)
		ifFalse: [^ false].
		
	(self signalConnections at: aSignal)
		removeKey: aReceiver
		ifAbsent: [^ false].
	
	(self signalConnections at: aSignal)
		ifEmpty: [^ self disconnectSignal: aSignal].
		
	^ true! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/2/2010 22:13'!
disconnectSignal: aSignal from: aReceiver selector: aSelector
	
	(self signalConnections includesKey: aSignal)
		ifFalse: [^ false].
		
	((self signalConnections at: aSignal)
		at: aReceiver
		ifAbsent: [^ false]) "weak"
		removeAllSuchThat: [:selectorAndPattern | selectorAndPattern key = aSelector].
	
	((self signalConnections at: aSignal)
		at: aReceiver
		ifAbsent: [^ true]) "weak"
		ifEmpty: [self disconnectSignal: aSignal from: aReceiver].
	
	^ true! !

!Object methodsFor: '*signals-core' stamp: 'mt 6/4/2010 12:55'!
emit
	"Should be called from signals only."

	| signalCtx signal args actions |
	signalCtx := thisContext sender.
	signal := signalCtx method.
	
	"Extract arguments."
	args := Array new: signal numArgs.
	(1 to: signal numArgs) do: [:index |
		args at: index put: (signalCtx tempAt: index)].
	
	"Fetch targets and dispatch."
	actions := OrderedCollection new. "Buffer actions to free the mutex in time."
	SignalConnectionsRepository critical: [
		(self signalConnections at: signal selector ifAbsent: [^ self]) associationsDo: [ :association |
			association key ifNotNil: [ :receiver | "Don't let the key go away -> weak dictionary"
				association value do: [ :selectorWithPatternQueue |
					| selector pattern queue filteredArgs blocking forceUiProcess |
					selector := selectorWithPatternQueue key.
					pattern := selectorWithPatternQueue value first.
					queue := selectorWithPatternQueue value second.
					blocking := selectorWithPatternQueue value third.
					forceUiProcess := selectorWithPatternQueue value fourth.
					
					forceUiProcess ifTrue: [
						blocking := false. "Never block here."
						queue := Processor activeProcess == Project uiProcess
							ifTrue: [nil] "synchron processing in UI process"
							ifFalse: [WorldState deferredUIMessages]]. "asynchron processing in UI process"
					
					pattern
						ifNil: [ "no argument changes"
							filteredArgs := args]
						ifNotNil: [ "shadowing and reordering"
							filteredArgs := Array new: pattern size.
							pattern withIndexDo: [:source :target | filteredArgs at: target put: (args at: source)]].
					actions add: (queue "queued connection or direct connection?"
						ifNil: [[receiver perform: selector withArguments: filteredArgs]]
						ifNotNil: [
							blocking
								ifTrue: [ | thisProcess |
									thisProcess := Processor activeProcess.
									[queue nextPut: [receiver perform: selector withArguments: filteredArgs. thisProcess resume].  thisProcess suspend]]
								ifFalse: [ [queue nextPut: [receiver perform: selector withArguments: filteredArgs]] ] ] ) ] ] ] ].
	
	actions do: [:action | action value].
	
	"TODO: Handle dispatching errors with automatic disconnections."
	! !

!Object methodsFor: '*signals-core-private' stamp: 'mt 6/2/2010 22:41'!
signalConnections

	^ SignalConnectionsRepository instance
		at: self
		ifAbsentPut: [Dictionary new]! !

Object subclass: #SignalConsumer
	instanceVariableNames: 'value switch'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Signals-Tests'!

!SignalConsumer methodsFor: 'slots' stamp: 'mt 6/2/2010 21:43'!
switchState

	^ switch ifNil: [switch := #off]! !

!SignalConsumer methodsFor: 'slots' stamp: 'mt 6/2/2010 21:44'!
toggleSwitch

	switch := self switchState = #off
		ifTrue: [#on]
		ifFalse: [#off].! !

!SignalConsumer methodsFor: 'slots' stamp: 'mt 6/2/2010 21:43'!
value

	^ value! !

!SignalConsumer methodsFor: 'slots' stamp: 'mt 6/2/2010 21:43'!
value: aValue

	value := aValue.! !

Object subclass: #SignalProducer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Signals-Tests'!

!SignalProducer methodsFor: 'signals' stamp: 'mt 6/2/2010 21:39'!
newSignalProduced

	self emit.! !

!SignalProducer methodsFor: 'signals' stamp: 'mt 6/2/2010 21:39'!
valueChanged: aValue

	self emit.! !

!SignalProducer methodsFor: 'signals' stamp: 'mt 6/2/2010 22:20'!
valueChanged: aValue and: anotherValue

	self emit.! !

Object subclass: #SignalSpy
	instanceVariableNames: 'queue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Signals-Core'!

!SignalSpy commentStamp: 'mt 6/3/2010 11:55' prior: 0!
Useful for testing signals.!

!SignalSpy class methodsFor: 'instance creation' stamp: 'mt 6/3/2010 12:24'!
onSender: aSender signal: aSignal

	^ self new
		addSender: aSender
		signal: aSignal! !

!SignalSpy methodsFor: 'spying' stamp: 'mt 6/4/2010 13:00'!
addSender: aSender signal: aSignal

	[self
		connect: aSender
		signal: aSignal
		toSelector: aSignal] forkAndWait. "Avoid processing of signals in UI process."! !

!SignalSpy methodsFor: 'spying' stamp: 'mt 6/3/2010 12:02'!
doesNotUnderstand: aMessage

	self queue nextPut: {aMessage selector. aMessage arguments}.! !

!SignalSpy methodsFor: 'spying' stamp: 'mt 6/3/2010 11:58'!
queue

	^ queue ifNil: [queue := SharedQueue new]! !

!SignalSpy methodsFor: 'spying' stamp: 'mt 6/3/2010 12:00'!
respondsTo: aSelector

	^ true! !

!SignalSpy methodsFor: 'waiting' stamp: 'mt 6/3/2010 12:07'!
waitForNextSignal

	^ self queue next! !

!SignalSpy methodsFor: 'waiting' stamp: 'mt 6/3/2010 12:21'!
waitForSignal: aSignal

	| signal |
	[signal notNil and: [signal first = aSignal]]
		whileFalse: [signal := self queue next].
		
	^ signal! !

WeakIdentityKeyDictionary subclass: #SignalConnectionsRepository
	instanceVariableNames: 'mutex'
	classVariableNames: 'SignalConnectionsRepository'
	poolDictionaries: ''
	category: 'Signals-Core'!

!SignalConnectionsRepository commentStamp: 'mt 6/2/2010 22:43' prior: 0!
I store all active signal-selector connections with weak data structures.!

!SignalConnectionsRepository class methodsFor: 'instance' stamp: 'mt 6/3/2010 08:50'!
critical: aBlock

	^ self instance critical: aBlock! !

!SignalConnectionsRepository class methodsFor: 'instance' stamp: 'mt 6/2/2010 21:28'!
instance

	^ SignalConnectionsRepository ifNil: [SignalConnectionsRepository := self new]! !

!SignalConnectionsRepository methodsFor: 'critical' stamp: 'mt 6/3/2010 08:46'!
critical: aBlock

	mutex ifNil: [mutex := Mutex new].
	^ mutex critical: aBlock! !
